from datetime import date, datetime, timezone

from functools import reduce
from operator import and_
from django.db.models import F, ManyToManyField, OuterRef, Subquery, Q, Prefetch

import graphene
from graphene_django import DjangoObjectType
from graphene_django_cud.mutations import DjangoBatchPatchMutation, DjangoCreateMutation, DjangoPatchMutation

from aleksis.core.models import CalendarEvent, Group, Person, Room
from aleksis.core.schema.base import DeleteMutation

from aleksis.apps.kolego.models import Absence

from ..models import LessonEvent
from ..util.chronos_helpers import get_classes, get_rooms, get_teachers


class TimetablePersonType(DjangoObjectType):
    class Meta:
        model = Person
        fields = ("id", "first_name", "last_name", "short_name")
        skip_registry = True


class TimetableGroupType(DjangoObjectType):
    class Meta:
        model = Group
        fields = ("id", "name", "short_name")
        skip_registry = True


class TimetableRoomType(DjangoObjectType):
    class Meta:
        model = Room
        fields = ("id", "name", "short_name")
        skip_registry = True


# There is another unrelated CalendarEventType in aleksis/core/schema/calendar
# This CalendarEventType is needed for the inherited amends field of LessonEvent
# to work in the graphql query.
class CalendarEventForLessonEventType(DjangoObjectType):
    class Meta:
        model = CalendarEvent
        fields = ("id", "amends", "datetime_start", "datetime_end")


class LessonEventType(DjangoObjectType):
    class Meta:
        model = LessonEvent
        fields = (
            "id",
            "amends",
            "datetime_start",
            "datetime_end",
            "subject",
            "teachers",
            "groups",
            "rooms",
            "course",
            "cancelled",
            "comment",
        )


class LessonEventTypeWithRealAmends(DjangoObjectType):
    class Meta:
        model = LessonEvent
        fields = (
            "id",
            "amends",
            "datetime_start",
            "datetime_end",
            "subject",
            "teachers",
            "groups",
            "rooms",
            "course",
            "cancelled",
            "comment",
        )

    real_amends = graphene.Field(LessonEventType, required=False)


class DatetimeTimezoneMixin:
    """Handle datetimes for mutations with CalendarEvent objects.

    This is necessary because the client sends timezone information as
    ISO string which only includes an offset (+00:00 UTC) and an
    offset is not a valid timezone. Instead we set UTC as timezone
    here directly.
    """

    @classmethod
    def handle_datetime_start(cls, value, name, info) -> int:
        value = value.replace(tzinfo=timezone.utc)
        return value

    @classmethod
    def handle_datetime_end(cls, value, name, info) -> int:
        value = value.replace(tzinfo=timezone.utc)
        return value

    @classmethod
    def before_save(cls, root, info, input, obj, patch_obj=False):
        # before_save has different signatures for different mutations
        # This handles create & patch
        # https://graphene-django-cud.readthedocs.io/en/latest/guide/other-hooks.html?highlight=before_save#before-save

        if patch_obj:
            obj = patch_obj
            
        obj.timezone = obj.amends.timezone
        return obj


class AmendLessonCreateMutation(DatetimeTimezoneMixin, DjangoCreateMutation):
    class Meta:
        model = LessonEvent
        permissions = ("chronos.edit_substitution_rule",)
        only_fields = (
            "amends",
            "datetime_start",
            "datetime_end",
            "subject",
            "teachers",
            "groups",
            "rooms",
            "cancelled",
            "comment",
        )


class AmendLessonPatchMutation(DatetimeTimezoneMixin, DjangoPatchMutation):
    class Meta:
        model = LessonEvent
        permissions = ("chronos.edit_substitution_rule",)
        only_fields = ("subject", "teachers", "groups", "rooms", "cancelled", "comment")


class AmendLessonBatchPatchMutation(DjangoBatchPatchMutation):
    class Meta:
        model = LessonEvent
        permissions = ("chronos.edit_substitution_rule",)
        only_fields = ("id", "subject", "teachers", "groups", "rooms", "cancelled", "comment")


class AmendLessonDeleteMutation(DeleteMutation):
    klass = LessonEvent
    permission_required = "chronos.edit_substitution_rule"


class TimetableType(graphene.Enum):
    TEACHER = "teacher"
    GROUP = "group"
    ROOM = "room"


class TimetableObjectType(graphene.ObjectType):
    id = graphene.String()  # noqa
    obj_id = graphene.String()
    name = graphene.String()
    short_name = graphene.String()
    type = graphene.Field(TimetableType)  # noqa

    def resolve_obj_id(root, info, **kwargs):
        return root.id

    def resolve_id(root, info, **kwargs):
        return f"{root.type.value}-{root.id}"


class LessonEventType(DjangoObjectType):
    class Meta:
        model = LessonEvent
        fields = ("id",)
        filter_fields = {
            "id": ["exact", "lte", "gte"],
        }


class Query(graphene.ObjectType):
    timetable_teachers = graphene.List(TimetablePersonType)
    timetable_groups = graphene.List(TimetableGroupType)
    timetable_rooms = graphene.List(TimetableRoomType)
    available_timetables = graphene.List(TimetableObjectType)

    amended_lessons_from_absences = graphene.List(
        LessonEventTypeWithRealAmends,
        filters=graphene.JSONString(required=True),
    )

    def resolve_timetable_teachers(self, info, **kwargs):
        return get_teachers(info.context.user)

    def resolve_timetable_groups(self, info, **kwargs):
        return get_classes(info.context.user)

    def resolve_timetable_rooms(self, info, **kwargs):
        return get_rooms(info.context.user)

    def resolve_available_timetables(self, info, **kwargs):
        all_timetables = []
        for group in get_classes(info.context.user):
            all_timetables.append(
                TimetableObjectType(
                    id=group.id,
                    name=group.name,
                    short_name=group.short_name,
                    type=TimetableType.GROUP,
                )
            )

        for teacher in get_teachers(info.context.user):
            print(teacher.full_name)
            all_timetables.append(
                TimetableObjectType(
                    id=teacher.id,
                    name=teacher.full_name,
                    short_name=teacher.short_name,
                    type=TimetableType.TEACHER,
                )
            )

        for room in get_rooms(info.context.user):
            all_timetables.append(
                TimetableObjectType(
                    id=room.id, name=room.name, short_name=room.short_name, type=TimetableType.ROOM
                )
            )

        return all_timetables

    def resolve_amended_lessons_from_absences(root, info, filters, **kwargs):
        if isinstance(filters, str):
            filters = json.loads(filters)

        datetime_start = datetime.combine(date.fromisoformat(filters.get("date_start", datetime.now().date().isoformat())) , datetime.min.time())
        datetime_end = datetime.combine(date.fromisoformat(filters.get("date_end", datetime.now().date().isoformat())), datetime.max.time())

        group_id = filters.get("group_id")

        # TODO: later on, allow getting amended lessons for other types than courses, e.g. groups or persons
        return LessonEvent.get_for_substitution_overview("GROUP", group_id, datetime_start, datetime_end, info.context)


class Mutation(graphene.ObjectType):
    create_amend_lesson = AmendLessonCreateMutation.Field()
    patch_amend_lesson = AmendLessonPatchMutation.Field()
    batch_patch_amend_lessons = AmendLessonBatchPatchMutation.Field()
    delete_amend_lesson = AmendLessonDeleteMutation.Field()
